# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
from os.path import abspath
from struct import unpack
from ev3dev2 import get_current_platform, Device, list_device_names

# INPUT ports have platform specific values that we must import
platform = get_current_platform()

if platform == 'ev3':
    from ev3dev2._platform.ev3 import INPUT_1, INPUT_2, INPUT_3, INPUT_4  # noqa: F401

elif platform == 'evb':
    from ev3dev2._platform.evb import INPUT_1, INPUT_2, INPUT_3, INPUT_4  # noqa: F401

elif platform == 'pistorms':
    from ev3dev2._platform.pistorms import INPUT_1, INPUT_2, INPUT_3, INPUT_4  # noqa: F401

elif platform == 'brickpi':
    from ev3dev2._platform.brickpi import INPUT_1, INPUT_2, INPUT_3, INPUT_4  # noqa: F401

elif platform == 'brickpi3':
    from ev3dev2._platform.brickpi3 import (  # noqa: F401
        INPUT_1, INPUT_2, INPUT_3, INPUT_4, INPUT_5, INPUT_6, INPUT_7, INPUT_8, INPUT_9, INPUT_10, INPUT_11, INPUT_12,
        INPUT_13, INPUT_14, INPUT_15, INPUT_16)

elif platform == 'fake':
    from ev3dev2._platform.fake import INPUT_1, INPUT_2, INPUT_3, INPUT_4  # noqa: F401

else:
    raise Exception("Unsupported platform '%s'" % platform)

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')


class Sensor(Device):
    """
    The sensor class provides a uniform interface for using most of the
    sensors available for the EV3.
    """

    SYSTEM_CLASS_NAME = 'lego-sensor'
    SYSTEM_DEVICE_NAME_CONVENTION = 'sensor*'
    __slots__ = [
        '_address', '_command', '_commands', '_decimals', '_driver_name', '_mode', '_modes', '_num_values', '_units',
        '_value', '_bin_data_format', '_bin_data_size', '_bin_data', '_mode_scale'
    ]

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):

        if address is not None:
            kwargs['address'] = address
        super(Sensor, self).__init__(self.SYSTEM_CLASS_NAME, name_pattern, name_exact, **kwargs)

        self._address = None
        self._command = None
        self._commands = None
        self._decimals = None
        self._driver_name = None
        self._mode = None
        self._modes = None
        self._num_values = None
        self._units = None
        self._value = [None, None, None, None, None, None, None, None]

        self._bin_data_format = None
        self._bin_data_size = None
        self._bin_data = None
        self._mode_scale = {}

    def _scale(self, mode):
        """
        Returns value scaling coefficient for the given mode.
        """
        if mode in self._mode_scale:
            scale = self._mode_scale[mode]
        else:
            scale = 10**(-self.decimals)
            self._mode_scale[mode] = scale

        return scale

    @property
    def address(self):
        """
        Returns the name of the port that the sensor is connected to, e.g. ``ev3:in1``.
        I2C sensors also include the I2C address (decimal), e.g. ``ev3:in1:i2c8``.
        """
        self._address, value = self.get_attr_string(self._address, 'address')
        return value

    @property
    def command(self):
        """
        Sends a command to the sensor.
        """
        raise Exception("command is a write-only property!")

    @command.setter
    def command(self, value):
        self._command = self.set_attr_string(self._command, 'command', value)

    @property
    def commands(self):
        """
        Returns a list of the valid commands for the sensor.
        Returns -EOPNOTSUPP if no commands are supported.
        """
        (self._commands, value) = self.get_cached_attr_set(self._commands, 'commands')
        return value

    @property
    def decimals(self):
        """
        Returns the number of decimal places for the values in the ``value<N>``
        attributes of the current mode.
        """
        self._decimals, value = self.get_attr_int(self._decimals, 'decimals')
        return value

    @property
    def driver_name(self):
        """
        Returns the name of the sensor device/driver. See the list of [supported
        sensors] for a complete list of drivers.
        """
        (self._driver_name, value) = self.get_cached_attr_string(self._driver_name, 'driver_name')
        return value

    @property
    def mode(self):
        """
        Returns the current mode. Writing one of the values returned by ``modes``
        sets the sensor to that mode.
        """
        self._mode, value = self.get_attr_string(self._mode, 'mode')
        return value

    @mode.setter
    def mode(self, value):
        self._mode = self.set_attr_string(self._mode, 'mode', value)

    @property
    def modes(self):
        """
        Returns a list of the valid modes for the sensor.
        """
        (self._modes, value) = self.get_cached_attr_set(self._modes, 'modes')
        return value

    @property
    def num_values(self):
        """
        Returns the number of ``value<N>`` attributes that will return a valid value
        for the current mode.
        """
        self._num_values, value = self.get_attr_int(self._num_values, 'num_values')
        return value

    @property
    def units(self):
        """
        Returns the units of the measured value for the current mode. May return
        empty string
        """
        self._units, value = self.get_attr_string(self._units, 'units')
        return value

    def value(self, n=0):
        """
        Returns the value or values measured by the sensor. Check num_values to
        see how many values there are. Values with N >= num_values will return
        an error. The values are fixed point numbers, so check decimals to see
        if you need to divide to get the actual value.
        """
        n = int(n)

        self._value[n], value = self.get_attr_int(self._value[n], 'value' + str(n))
        return value

    @property
    def bin_data_format(self):
        """
        Returns the format of the values in ``bin_data`` for the current mode.
        Possible values are:

        - ``u8``: Unsigned 8-bit integer (byte)
        - ``s8``: Signed 8-bit integer (sbyte)
        - ``u16``: Unsigned 16-bit integer (ushort)
        - ``s16``: Signed 16-bit integer (short)
        - ``s16_be``: Signed 16-bit integer, big endian
        - ``s32``: Signed 32-bit integer (int)
        - ``float``: IEEE 754 32-bit floating point (float)
        """
        self._bin_data_format, value = self.get_attr_string(self._bin_data_format, 'bin_data_format')
        return value

    def bin_data(self, fmt=None):
        """
        Returns the unscaled raw values in the ``value<N>`` attributes as raw byte
        array. Use ``bin_data_format``, ``num_values`` and the individual sensor
        documentation to determine how to interpret the data.

        Use ``fmt`` to unpack the raw bytes into a struct.

        Example::

            >>> from ev3dev2.sensor.lego import InfraredSensor
            >>> ir = InfraredSensor()
            >>> ir.value()
            28
            >>> ir.bin_data('<b')
            (28,)
        """

        if self._bin_data_size is None:
            self._bin_data_size = {
                "u8": 1,
                "s8": 1,
                "u16": 2,
                "s16": 2,
                "s16_be": 2,
                "s32": 4,
                "float": 4
            }.get(self.bin_data_format, 1) * self.num_values

        if self._bin_data is None:
            self._bin_data = self._attribute_file_open('bin_data')

        self._bin_data.seek(0)
        raw = bytearray(self._bin_data.read(self._bin_data_size))

        if fmt is None:
            return raw

        return unpack(fmt, raw)

    def _ensure_mode(self, mode):
        if self.mode != mode:
            self.mode = mode


def list_sensors(name_pattern=Sensor.SYSTEM_DEVICE_NAME_CONVENTION, **kwargs):
    """
    This is a generator function that enumerates all sensors that match the
    provided arguments.

    Parameters:
        name_pattern: pattern that device name should match.
            For example, 'sensor*'. Default value: '*'.
        keyword arguments: used for matching the corresponding device
            attributes. For example, driver_name='lego-ev3-touch', or
            address=['in1', 'in3']. When argument value is a list,
            then a match against any entry of the list is enough.
    """
    class_path = abspath(Device.DEVICE_ROOT_PATH + '/' + Sensor.SYSTEM_CLASS_NAME)
    return (Sensor(name_pattern=name, name_exact=True)
            for name in list_device_names(class_path, name_pattern, **kwargs))


class I2cSensor(Sensor):
    """
    A generic interface to control I2C-type EV3 sensors.
    """

    SYSTEM_CLASS_NAME = Sensor.SYSTEM_CLASS_NAME
    SYSTEM_DEVICE_NAME_CONVENTION = 'sensor*'

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):
        super(I2cSensor, self).__init__(address, name_pattern, name_exact, driver_name=['nxt-i2c-sensor'], **kwargs)
        self._fw_version = None
        self._poll_ms = None

    @property
    def fw_version(self):
        """
        Returns the firmware version of the sensor if available. Currently only
        I2C/NXT sensors support this.
        """
        (self._fw_version, value) = self.get_cached_attr_string(self._fw_version, 'fw_version')
        return value

    @property
    def poll_ms(self):
        """
        Returns the polling period of the sensor in milliseconds. Writing sets the
        polling period. Setting to 0 disables polling. Minimum value is hard
        coded as 50 msec. Returns -EOPNOTSUPP if changing polling is not supported.
        Currently only I2C/NXT sensors support changing the polling period.
        """
        self._poll_ms, value = self.get_attr_int(self._poll_ms, 'poll_ms')
        return value

    @poll_ms.setter
    def poll_ms(self, value):
        self._poll_ms = self.set_attr_int(self._poll_ms, 'poll_ms', value)



from mindpile.Mapping.utils import MethodCall, Requires, Setup

# TODO: implement playing files, non-blocking looping playtype, stopping sound once there's support in ev3dev

@Setup
def soundSetup():
    return '''
        from ev3dev2.sound import Sound
        speaker = Sound()
    '''

@Setup
def soundStopNotImplemented():
    return '''
        print("WARNING: Stop Sound is not implemented")
    '''

@MethodCall(target="PlaySoundStop.vix")
@Requires(soundSetup)
@Requires(soundStopNotImplemented)
def soundStop():
    return '''
    '''

@Setup
def soundFileNotImplemented():
    return '''
        print("WARNING: Play Sound File is not implemented")
    '''

@MethodCall(target="PlaySoundFile.vix")
@Requires(soundSetup)
@Requires(soundFileNotImplemented)
def soundFile():
    return '''
    '''

@MethodCall(target="PlayTone.vix", Frequency=float, Duration=float, Volume=int, PlayType=int)
@Requires(soundSetup)
def soundTone():
    return '''
        speaker.play_tone(frequency=Frequency, duration=Duration, volume=Volume, play_type=PlayType)
    '''

@MethodCall(target="PlayNote.vix", Note=str, Duration=float, Volume=int, PlayType=int)
@Requires(soundSetup)
def soundNote():
    return '''
        speaker.play_note(note=Note, duration=Duration, volume=Volume, play_type=PlayType)
    '''



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
from . import Device

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')


class LegoPort(Device):
    """
    The ``lego-port`` class provides an interface for working with input and
    output ports that are compatible with LEGO MINDSTORMS RCX/NXT/EV3, LEGO
    WeDo and LEGO Power Functions sensors and motors. Supported devices include
    the LEGO MINDSTORMS EV3 Intelligent Brick, the LEGO WeDo USB hub and
    various sensor multiplexers from 3rd party manufacturers.

    See the following example for using this class to configure sensors:
    https://github.com/ev3dev/ev3dev-lang-python-demo/blob/stretch/platform/brickpi3-motor-and-sensor.py

    Some types of ports may have multiple modes of operation. For example, the
    input ports on the EV3 brick can communicate with sensors using UART, I2C
    or analog validate signals - but not all at the same time. Therefore there
    are multiple modes available to connect to the different types of sensors.

    In most cases, ports are able to automatically detect what type of sensor
    or motor is connected. In some cases though, this must be manually specified
    using the ``mode`` and ``set_device`` attributes. The ``mode`` attribute affects
    how the port communicates with the connected device. For example the input
    ports on the EV3 brick can communicate using UART, I2C or analog voltages,
    but not all at the same time, so the mode must be set to the one that is
    appropriate for the connected sensor. The ``set_device`` attribute is used to
    specify the exact type of sensor that is connected. Note: the mode must be
    correctly set before setting the sensor type.

    Ports can be found at ``/sys/class/lego-port/port<N>`` where ``<N>`` is
    incremented each time a new port is registered. Note: The number is not
    related to the actual port at all - use the ``address`` attribute to find
    a specific port.
    """

    SYSTEM_CLASS_NAME = 'lego-port'
    SYSTEM_DEVICE_NAME_CONVENTION = '*'
    __slots__ = [
        '_address',
        '_driver_name',
        '_modes',
        '_mode',
        '_set_device',
        '_status',
    ]

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):

        if address is not None:
            kwargs['address'] = address
        super(LegoPort, self).__init__(self.SYSTEM_CLASS_NAME, name_pattern, name_exact, **kwargs)

        self._address = None
        self._driver_name = None
        self._modes = None
        self._mode = None
        self._set_device = None
        self._status = None

    @property
    def address(self):
        """
        Returns the name of the port. See individual driver documentation for
        the name that will be returned.
        """
        self._address, value = self.get_attr_string(self._address, 'address')
        return value

    @property
    def driver_name(self):
        """
        Returns the name of the driver that loaded this device. You can find the
        complete list of drivers in the [list of port drivers].
        """
        (self._driver_name, value) = self.get_cached_attr_string(self._driver_name, 'driver_name')
        return value

    @property
    def modes(self):
        """
        Returns a list of the available modes of the port.
        """
        (self._modes, value) = self.get_cached_attr_set(self._modes, 'modes')
        return value

    @property
    def mode(self):
        """
        Reading returns the currently selected mode. Writing sets the mode.
        Generally speaking when the mode changes any sensor or motor devices
        associated with the port will be removed new ones loaded, however this
        this will depend on the individual driver implementing this class.
        """
        self._mode, value = self.get_attr_string(self._mode, 'mode')
        return value

    @mode.setter
    def mode(self, value):
        self._mode = self.set_attr_string(self._mode, 'mode', value)

    @property
    def set_device(self):
        """
        For modes that support it, writing the name of a driver will cause a new
        device to be registered for that driver and attached to this port. For
        example, since NXT/Analog sensors cannot be auto-detected, you must use
        this attribute to load the correct driver. Returns -EOPNOTSUPP if setting a
        device is not supported.
        """
        raise Exception("set_device is a write-only property!")

    @set_device.setter
    def set_device(self, value):
        self._set_device = self.set_attr_string(self._set_device, 'set_device', value)

    @property
    def status(self):
        """
        In most cases, reading status will return the same value as ``mode``. In
        cases where there is an ``auto`` mode additional values may be returned,
        such as ``no-device`` or ``error``. See individual port driver documentation
        for the full list of possible values.
        """
        self._status, value = self.get_attr_string(self._status, 'status')
        return value



from collections import OrderedDict

# Up to four brickpi3s can be stacked
OUTPUT_A = 'spi0.1:MA'
OUTPUT_B = 'spi0.1:MB'
OUTPUT_C = 'spi0.1:MC'
OUTPUT_D = 'spi0.1:MD'

OUTPUT_E = 'spi0.1:ME'
OUTPUT_F = 'spi0.1:MF'
OUTPUT_G = 'spi0.1:MG'
OUTPUT_H = 'spi0.1:MH'

OUTPUT_I = 'spi0.1:MI'
OUTPUT_J = 'spi0.1:MJ'
OUTPUT_K = 'spi0.1:MK'
OUTPUT_L = 'spi0.1:ML'

OUTPUT_M = 'spi0.1:MM'
OUTPUT_N = 'spi0.1:MN'
OUTPUT_O = 'spi0.1:MO'
OUTPUT_P = 'spi0.1:MP'

INPUT_1 = 'spi0.1:S1'
INPUT_2 = 'spi0.1:S2'
INPUT_3 = 'spi0.1:S3'
INPUT_4 = 'spi0.1:S4'

INPUT_5 = 'spi0.1:S5'
INPUT_6 = 'spi0.1:S6'
INPUT_7 = 'spi0.1:S7'
INPUT_8 = 'spi0.1:S8'

INPUT_9 = 'spi0.1:S9'
INPUT_10 = 'spi0.1:S10'
INPUT_11 = 'spi0.1:S11'
INPUT_12 = 'spi0.1:S12'

INPUT_13 = 'spi0.1:S13'
INPUT_14 = 'spi0.1:S14'
INPUT_15 = 'spi0.1:S15'
INPUT_16 = 'spi0.1:S16'

BUTTONS_FILENAME = None
EVDEV_DEVICE_NAME = None

LEDS = OrderedDict()
LEDS['amber_led'] = 'led1:amber:brick-status'

LED_GROUPS = OrderedDict()
LED_GROUPS['LED'] = ('amber_led', )

LED_COLORS = OrderedDict()
LED_COLORS['BLACK'] = (0, )
LED_COLORS['AMBER'] = (1, )

LED_DEFAULT_COLOR = 'AMBER'



from mindpile.Mapping.types import OutPort
from mindpile.Mapping.utils import MethodCall, Requires, Setup

@Setup
def mediumMotorSetup():
    return '''
        from ev3dev2.motor import MediumMotor
    '''

@MethodCall(target="MediumMotorStop.vix", MotorPort=OutPort, BrakeAtEnd=bool)
@Requires(mediumMotorSetup)
def mediumMotorStop():
    return '''
        m(MotorPort, MediumMotor).stop(stop_action=("hold" if BrakeAtEnd else "coast"))
    '''

@MethodCall(target="MediumMotorUnlimited.vix", MotorPort=OutPort, Speed=float)
@Requires(mediumMotorSetup)
def mediumMotorUnlimited():
    return '''
        m(MotorPort, MediumMotor).on(speed=Speed)
    '''

@MethodCall(target="MediumMotorTime.vix", MotorPort=OutPort, Speed=float, Seconds=float, BrakeAtEnd=bool)
@Requires(mediumMotorSetup)
def mediumMotorSeconds():
    return '''
        m(MotorPort, MediumMotor).on_for_seconds(speed=Speed, seconds=Seconds, brake=BrakeAtEnd)
    '''

@MethodCall(target="MediumMotorDistance.vix", MotorPort=OutPort, Speed=float, Degrees=float, BrakeAtEnd=bool)
@Requires(mediumMotorSetup)
def mediumMotorDegrees():
    return '''
        m(MotorPort, MediumMotor).on_for_degrees(speed=Speed, degrees=Degrees, brake=BrakeAtEnd)
    '''

@MethodCall(target="MediumMotorDistanceRotations.vix", MotorPort=OutPort, Speed=float, Rotations=float, BrakeAtEnd=bool)
@Requires(mediumMotorSetup)
def mediumMotorRotations():
    return '''
        m(MotorPort, MediumMotor).on_for_rotations(speed=Speed, rotations=Rotations, brake=BrakeAtEnd)
    '''



from typing import Tuple, List, Optional, Union
import mindpile.IntermediateFormat as IntermediateFormat
from mindpile.Utility.types import *
from mindpile.Utility import utility as utility
from mindpile.Utility import exceptions as exceptions


def getSequenceTerminalInfo(elem: Terminal) -> Tuple[str, Optional[str]]:
    """
    Returns the type of sequence/flow terminal and the wire it attaches to
    """
    wire = None
    if "Wire" in elem.attrib:
        wire = elem.attrib["Wire"]
    if elem.attrib["Id"] == "SequenceIn" or elem.attrib["Direction"] == "Input":
        return "in", wire
    elif elem.attrib["Id"] == "SequenceOut" or elem.attrib["Direction"] == "Output":
        return "out", wire
    else:
        raise NotImplementedError(
            "Encountered Unexpected Bare terminal in method call", elem
        )


def getTerminalsFromParentElem(elem) -> Tuple[Optional[str],Optional[str]]:
    seqIn,seqOut = None,None
    for child in elem:
        tag = utility.removeNameSpace(child.tag)
        if tag == "Terminal":
            direction, wire = getSequenceTerminalInfo(child)
            if direction == "in":
                seqIn = wire
            else:
                seqOut = wire
    if seqIn == None and seqOut == None:
        raise exceptions.InvalidSourceFormatError("Expected a Terminal where none existed")
    return seqIn,seqOut

def whileLoop(elem: WhileLoop) -> IntermediateFormat.SequenceBlock:
    childBlocks: List[IntermediateFormat.SequenceBlock] = []
    seqIn: Optional[str] = None
    seqOut: Optional[str] = None
    seqIn, seqOut = getTerminalsFromParentElem(elem)
    for child in elem:
        tag = utility.removeNameSpace(child.tag)
        if tag == "ConfigurableWhileLoop.BuiltInMethod":
            if child.attrib["CallType"] == "LoopIndex":
                childBlocks.append(methodCall(child[0]))
            else:
                childBlocks.append(
                    methodCall(child[0], isBreakMethod=True)
                )  # normally we can just use the generic translater, but here we want to make sure the break method is known to be a break method, so we bypass that to pass the flag
        else:
            childBlocks.append(translateElementToIRForm(child))

    label = elem.attrib["InterruptName"]
    loop = IntermediateFormat.WhileLoop(label, childBlocks)
    return IntermediateFormat.SequenceBlock(elem.attrib["Id"], seqIn, seqOut, loop)


def configureableMethodTerminal(
    elem: ConfigurableMethodTerminal,
) -> Union[IntermediateFormat.Argument, IntermediateFormat.Output]:

    terminal = elem[0]
    name = terminal.attrib["Id"]
    dataType = terminal.attrib["DataType"]
    if terminal.attrib["Direction"] == "Output":
        if "Wire" in terminal.attrib:
            return IntermediateFormat.Output(name, dataType, terminal.attrib["Wire"])
        else:
            return IntermediateFormat.Output(name, dataType, None)
    else:
        if "Wire" in terminal.attrib:
            return IntermediateFormat.Argument(
                name, dataType, variableName=terminal.attrib["Wire"]
            )
        else:
            return IntermediateFormat.Argument(
                name, dataType, constValue=elem.attrib["ConfiguredValue"]
            )


def methodCall(
    elem: MethodCall, isBreakMethod=False, isWaitForMethod=False
) -> IntermediateFormat.SequenceBlock:

    functionName = elem.attrib["Target"]

    seqInputWire = None
    seqOutputWire = None
    arguments = []
    outputs = []
    for IO in elem:
        tag = utility.removeNameSpace(IO.tag)
        if tag == "ConfigurableMethodTerminal":
            if IO[0].attrib["Direction"] == "Input":
                arguments.append(configureableMethodTerminal(IO))
            else:  # is outpu [0].attrib["Direction"] == "Output"t
                outputs.append(configureableMethodTerminal(IO))
        elif tag == "Terminal":
            dir, wire = getSequenceTerminalInfo(IO)
            if dir == "in":
                seqInputWire = wire
            else:
                seqOutputWire = wire
        else:
            raise NotImplementedError("Unexpected tag in method call")
    if isBreakMethod:
        method = IntermediateFormat.BreakMethodCall(functionName, arguments, outputs)
    elif isWaitForMethod:
        method = IntermediateFormat.WaitForMethodCall(functionName, arguments, outputs)
    else:
        method = IntermediateFormat.MethodCall(functionName, arguments, outputs)

    return IntermediateFormat.SequenceBlock(
        elem.attrib["Id"], seqInputWire, seqOutputWire, method
    )

def configurableWaitForMethodCall(elem: MethodCall):
    return methodCall(elem, isWaitForMethod=True)

# THIS IS HARDCODED - unsure if this will prove problematic
def startBlock(elem):
    outputWire = elem[1].attrib["Wire"]  # this is the terminal output
    return IntermediateFormat.SequenceBlock(elem.attrib["Id"], None, outputWire, None)


def blockDiagram(elem):
    return IntermediateFormat.BlockDiagram(
        elem.attrib["Name"], [translateElementToIRForm(child) for child in elem]
    )


def configurableFlatCaseStructure(elem):
    seqIn, seqOut = getTerminalsFromParentElem(elem)

    cases = []
    for child in elem:
        tag = utility.removeNameSpace(child.tag)
        if tag == "ConfigurableFlatCaseStructure.Case":
            caseChildren = []
            for child2 in child:
                caseChildren.append(translateElementToIRForm(child2))
            newCase = IntermediateFormat.Case(child.attrib["Id"],child.attrib["Pattern"],caseChildren)
            cases.append(newCase)
    defaultCaseId = elem.attrib["Default"]
    switch = IntermediateFormat.SwitchCase(
        elem.attrib["Id"],
        elem.attrib["DataType"],
        elem.attrib["PairedConfigurableMethodCall"],
        cases,
        defaultCaseId
    )
    return IntermediateFormat.SequenceBlock(elem.attrib["Id"],seqIn,seqOut,switch)

#ok
#so basically
#we enter the paired method
#the value is computed
#then we enter the switch
#and that happens /IMPLICITLY/

def pairedMethodCall(elem:PairedConfigurableMethodCall):
    methodBlock = methodCall(elem) #reuse logic from method call, then extract data into proper sequenceblock
    innerMethod = methodBlock.logic
    pairedMethod = IntermediateFormat.PairedMethodCall(innerMethod,elem.attrib["PairedStructure"])

    return IntermediateFormat.SequenceBlock(methodBlock.id,methodBlock.inputWire,methodBlock.outputWire,pairedMethod)


def sequenceNode(elem:SequenceNode):
    seqIn, seqOut = getTerminalsFromParentElem(elem)
    return IntermediateFormat.SequenceBlock(elem.attrib["Id"],seqIn,seqOut,None)

def translateElementToIRForm(elem):
    tagToIRFunc = {
        "ConfigurableMethodCall": methodCall,
        "ConfigurableWhileLoop": whileLoop,
        "StartBlock": startBlock,
        "BlockDiagram": blockDiagram,
        "ConfigurableWaitFor": configurableWaitForMethodCall,
        "PairedConfigurableMethodCall": pairedMethodCall,  # TODO check this probably needs special handling for the paired swich
        "ConfigurableFlatCaseStructure": configurableFlatCaseStructure,
        "Terminal": lambda x: "Terminal??",  # these need to be handled appropriately
        "ConfigurableWhileLoop.BuiltInMethod": lambda x: "",
        "SequenceNode": sequenceNode,
        "Wire": lambda x: IntermediateFormat.SequenceBlock(
            "DEBUG NOT USED", "DEBUG NOT USED", "DEBUG DONT USE", None
        ),
    }
    return tagToIRFunc[utility.removeNameSpace(elem.tag)](elem)




from mindpile.Mapping.utils import MethodCall

@MethodCall(target="RandomSingle.vix", Lower=float, Upper=float, Number=float)
def randomNumeric():
    return '''
        Number = random.uniform(Lower, Upper)
    '''

@MethodCall(target="RandomBoolean.vix", PercentTrue=float, Result=bool)
def randomBoolean():
    return '''
        Result = random.random() * 100 <= PercentTrue
    '''



"""
A StopWatch class for tracking the amount of time between events
"""

from ev3dev2 import is_micropython

if is_micropython():
    import utime
else:
    import datetime as dt


def get_ticks_ms():
    if is_micropython():
        return utime.ticks_ms()
    else:
        return int(dt.datetime.timestamp(dt.datetime.now()) * 1000)


class StopWatchAlreadyStartedException(Exception):
    """
    Exception raised when start() is called on a StopWatch which was already start()ed and not yet
    stopped.
    """
    pass


class StopWatch(object):
    """
    A timer class which lets you start timing and then check the amount of time
    elapsed.
    """
    def __init__(self, desc=None):
        """
        Initializes the StopWatch but does not start it.

        desc:
            A string description to print when stringifying.
        """
        self.desc = desc
        self._start_time = None
        self._stopped_total_time = None

    def __str__(self):
        name = self.desc if self.desc is not None else self.__class__.__name__
        return "{}: {}".format(name, self.hms_str)

    def start(self):
        """
        Starts the timer. If the timer is already running, resets it.

        Raises a :py:class:`ev3dev2.stopwatch.StopWatchAlreadyStartedException` if already started.
        """
        if self.is_started:
            raise StopWatchAlreadyStartedException()

        self._stopped_total_time = None
        self._start_time = get_ticks_ms()

    def stop(self):
        """
        Stops the timer. The time value of this Stopwatch is paused and will not continue increasing.
        """
        if self._start_time is None:
            return

        self._stopped_total_time = get_ticks_ms() - self._start_time
        self._start_time = None

    def reset(self):
        """
        Resets the timer and leaves it stopped.
        """
        self._start_time = None
        self._stopped_total_time = None

    def restart(self):
        """
        Resets and then starts the timer.
        """
        self.reset()
        self.start()

    @property
    def is_started(self):
        """
        True if the StopWatch has been started but not stoped (i.e., it's currently running), false otherwise.
        """
        return self._start_time is not None

    @property
    def value_ms(self):
        """
        Returns the value of the stopwatch in milliseconds
        """
        if self._stopped_total_time is not None:
            return self._stopped_total_time

        return get_ticks_ms() - self._start_time if self._start_time is not None else 0

    @property
    def value_secs(self):
        """
        Returns the value of the stopwatch in seconds
        """
        return self.value_ms / 1000

    @property
    def value_hms(self):
        """
        Returns this StopWatch's elapsed time as a tuple
        ``(hours, minutes, seconds, milliseconds)``.
        """
        (hours, x) = divmod(int(self.value_ms), 3600000)
        (mins, x) = divmod(x, 60000)
        (secs, x) = divmod(x, 1000)
        return hours, mins, secs, x

    @property
    def hms_str(self):
        """
        Returns the stringified value of the stopwatch in HH:MM:SS.msec format
        """
        return '%02d:%02d:%02d.%03d' % self.value_hms

    def is_elapsed_ms(self, duration_ms):
        """
        Returns True if this timer has measured at least ``duration_ms``
        milliseconds.
        Otherwise, returns False. If ``duration_ms`` is None, returns False.
        """

        return duration_ms is not None and self.value_ms >= duration_ms

    def is_elapsed_secs(self, duration_secs):
        """
        Returns True if this timer has measured at least ``duration_secs`` seconds.
        Otherwise, returns False. If ``duration_secs`` is None, returns False.
        """

        return duration_secs is not None and self.value_secs >= duration_secs



from . import control
from . import ev3sim
from .actions import display
from .actions import largeMotor
from .actions import light
from .actions import mediumMotor
from .actions import sound
from .actions import steering
from .actions import tank
from .advanced import comment
from .data import array
from .data import compare
from .data import constants
from .data import logic
from .data import maths
from .data import random
from .data import range
from .data import round
from .data import text
from .data import variables
from .sensors import brick
from .sensors import colour
from .sensors import ir
from .sensors import motor



from mindpile.Mapping.types import InPort, List, Local
from mindpile.Mapping.utils import MethodCall, Requires, Setup

@Setup
def irSetup():
    return '''
        from ev3dev2.sensor.lego import InfraredSensor
    '''

@MethodCall(target="IRProximity.vix", Port=InPort, Proximity=int)
@Requires(irSetup)
def irProximity():
    return '''
        Proximity = s(Port, InfraredSensor).proximity
    '''

@MethodCall(target="IRSeeker.vix", Port=InPort, Channel=int, Heading=int, Proximity=int, Valid=bool)
@Requires(irSetup)
def irSeeker():
    return '''
        Heading, Proximity = s(Port, InfraredSensor).heading_and_distance(Channel)
        Valid = Proximity is not None
        if not Valid:
            Proximity = 100
    '''

@MethodCall(target="IRRemote.vix", Port=InPort, Channel=int, Button=int)
@Requires(irSetup)
def irRemote():
    return '''
        temp = s(Port, InfraredSensor)
        temp._ensure_mode(InfraredSensor.MODE_IR_REMOTE)
        Button = temp.value(Channel - 1)
    '''

@MethodCall(target="IRProximityCompare.vix", Port=InPort, Comparison=int, Threshold=int, Result=bool, Proximity=int)
@Requires(irSetup)
def irProximityCompare():
    return '''
        Proximity = s(Port, InfraredSensor).proximity
        Result = compare(Comparison, Proximity, Threshold)
    '''

@MethodCall(target="IRTrackerCompareHeading.vix", Port=InPort, Channel=int, Comparison=int, HeadingThreshold=int, Result=bool, Heading=int)
@Requires(irSetup)
def irSeekerCompareHeading():
    return '''
        Heading = s(Port, InfraredSensor).heading(Channel)
        Result = compare(Comparison, Heading, HeadingThreshold)
    '''

@MethodCall(target="IRTrackerCompareProximity.vix", Port=InPort, Channel=int, Comparison=int, Threshold=int, Result=bool, Proximity=int)
@Requires(irSetup)
def irSeekerCompareDistance():
    return '''
        Proximity = s(Port, InfraredSensor).distance(Channel)
        if Proximity is None:
            Proximity = 100
            Result = False
        else:
            Result = compare(Comparison, Proximity, Threshold)
    '''

@MethodCall(target="IRRemoteCompare.vix", Port=InPort, Channel=int, SetofremotebuttonIDs=List(int), Result=bool, Button=int)
@Requires(irSetup)
def irRemote():
    return '''
        temp = s(Port, InfraredSensor)
        temp._ensure_mode(InfraredSensor.MODE_IR_REMOTE)
        Button = temp.value(Channel - 1)
        Result = Button in SetofremotebuttonIDs
    '''

@MethodCall(target="IRProximityChange.vix", Port=InPort, Direction=int, Amount=int, Result=bool, Proximity=int, initValue=Local(None))
@Requires(irSetup)
def irProximityChange():
    return '''
        Proximity = s(Port, InfraredSensor).proximity
        if initValue is None:
            initValue = Proximity
        Result = compareDirection(Direction, Amount, initValue, Proximity)
        if Result:
            initValue = None
    '''

@MethodCall(target="IRTrackerChangeHeading.vix", Port=InPort, Direction=int, HeadingAmount=int, Result=bool, Heading=int, initValue=Local(None))
@Requires(irSetup)
def irSeekerChangeHeading():
    return '''
        Heading = s(Port, InfraredSensor).heading(Channel)
        if initValue is None:
            initValue = Heading
        Result = compareDirection(Direction, Amount, initValue, Heading)
        if Result:
            initValue = None
    '''

@MethodCall(target="IRTrackerChangeProximity.vix", Port=InPort, Direction=int, Amount=int, Result=bool, Proximity=int, initValue=Local(None))
@Requires(irSetup)
def irSeekerChangeDistance():
    return '''
        Proximity = s(Port, InfraredSensor).distance(Channel)
        if Proximity is None:
            Proximity = 100
        if initValue is None:
            initValue = Proximity
        Result = compareDirection(Direction, Amount, initValue, Proximity)
        if Result:
            initValue = None
    '''

@MethodCall(target="IRRemoteChange.vix", Port=InPort, Channel=int, Result=bool, Button=int, initValue=Local(None))
@Requires(irSetup)
def irProximityChange():
    return '''
        temp = s(Port, InfraredSensor)
        temp._ensure_mode(InfraredSensor.MODE_IR_REMOTE)
        Button = temp.value(Channel - 1)
        if Button is None:
            initValue = Button
        Result = Button != initValue
        if Result:
            initValue = None
    '''



import ast
from mindpile.Mapping.types import Literal
from mindpile.Mapping.utils import MethodCall, DynamicMethodCall, Requires, Setup, p

@MethodCall(target="Arithmetic_Add.vix", X=float, Y=float, Result=float)
def add():
    return '''
        Result = X + Y
    '''

@MethodCall(target="Arithmetic_Subtract.vix", X=float, Y=float, Result=float)
def sub():
    return '''
        Result = X - Y
    '''

@MethodCall(target="Arithmetic_Divide.vix", X=float, Y=float, Result=float)
def div():
    return '''
        Result = X / Y
    '''

@MethodCall(target="Arithmetic_Multiply.vix", X=float, Y=float, Result=float)
def mul():
    return '''
        Result = X * Y
    '''

@MethodCall(target="Arithmetic_AbsoluteValue.vix", X=float, Result=float)
def absolute():
    return '''
        Result = abs(X)
    '''

@MethodCall(target="Arithmetic_SquareRoot.vix", X=float, Result=float)
def sqrt():
    return '''
        Result = math.sqrt(X)
    '''

@MethodCall(target="Arithmetic_Power.vix", Base=float, Exponent=float, Result=float)
def power():
    return '''
        Result = Base ** Exponent
    '''

@Setup
def advancedMathsSetup():
    return '''
        def sinDeg(x):
            return math.degrees(math.sin(x))
        def cosDeg(x):
            return math.degrees(math.cos(x))
        def tanDeg(x):
            return math.degrees(math.tan(x))
        def asinDeg(x):
            return math.degrees(math.asin(x))
        def acosDeg(x):
            return math.degrees(math.acos(x))
        def atanDeg(x):
            return math.degrees(math.atan(x))
    '''

@DynamicMethodCall(target="X3Placeholder_2A058539-ED76-4476-93FE-CCE8AA559C5A_MathEquation.vix", X=float, Y=float, C=float, D=float, Equation=Literal(str), Result=float)
@Requires(advancedMathsSetup)
def advancedMaths(X, Y, C, D, Equation, Result, **kwargs):
    tree = p(f'''
        Result = {Equation.lower()}
    ''', {
        "Result": Result,
        "a": X,
        "b": Y,
        "c": C,
        "d": D,
        "floor": ast.Attribute(value=ast.Name(id='math', ctx=ast.Load()), attr='floor', ctx=ast.Load()),
        "ceil": ast.Attribute(value=ast.Name(id='math', ctx=ast.Load()), attr='ceil', ctx=ast.Load()),
        "round": ast.Name(id="round", ctx=ast.Load()),
        "abs": ast.Name(id="abs", ctx=ast.Load()),
        # TODO: Check that log() is a log10 function
        "log": ast.Attribute(value=ast.Name(id='math', ctx=ast.Load()), attr='log10', ctx=ast.Load()),
        "ln": ast.Attribute(value=ast.Name(id='math', ctx=ast.Load()), attr='log', ctx=ast.Load()),
        "sin": ast.Name(id="sinDeg", ctx=ast.Load()),
        "cos": ast.Name(id="cosDeg", ctx=ast.Load()),
        "tan": ast.Name(id="tanDeg", ctx=ast.Load()),
        "asin": ast.Name(id="asinDeg", ctx=ast.Load()),
        "acos": ast.Name(id="acosDeg", ctx=ast.Load()),
        "atan": ast.Name(id="atanDeg", ctx=ast.Load()),
        "sqrt": ast.Attribute(value=ast.Name(id='math', ctx=ast.Load()), attr='sqrt', ctx=ast.Load()),
    })

    # Substitute the power (^) symbol with python's (**)
    # TODO: Ensure that power has highest order of operations in EV3 and it's same as python
    class Transformer(ast.NodeTransformer):
        def visit_BitXor(self, node: ast.BitXor):
            return ast.Pow()
    transformer = Transformer()
    transformer.visit(tree)

    return tree



from mindpile.Mapping.types import List
from mindpile.Mapping.utils import MethodCall, Requires, Setup

# TODO: verify button code works on brick before enabling it

@Setup
def buttonSetup():
    return '''
        import ev3dev2.button
        brickButtons = ev3dev2.button.Button()
        buttonMap = [None, "left", "enter", "right", "up", "down"]
        buttonMapR = { "left": 1, "enter": 2, "right": 3, "up": 4, "down": 5 }
        
        def buttonValue():
            pressed = brickButtons.buttons_pressed
            if len(pressed) == 0:
                return 0
            else:
                return buttonMapR[pressed[0]]
        
        buttonsBumped = [None, False, False, False, False, False]
        buttonsPressedTime = [None, None, None, None, None, None]
        buttonBumpTimeout = 0.5
        def buttonChanged(changedButtons):
            for button, state in changedButtons:
                b = buttonMapR[button]
                if state:
                    buttonsPressedTime[b] = time.time()
                else:
                    if buttonsPressedTime[b] is not None and time.time() <= buttonsPressedTime[b] + buttonBumpTimeout:
                        buttonsBumped[b] = True
                    buttonsPressedTime[b] = None
        brickButtons.on_change = buttonChanged

        def buttonCompare(buttons, action):
            buttonsPressed = brickButtons.buttons_pressed
            for b in buttons:
                button = buttonMap[b]
                if action == 0 and button not in buttonsPressed:
                    return True, b
                elif action == 1 and button in buttonsPressed:
                    return True, b
                elif action == 2 and buttonsBumped[b]:
                    buttonsBumped[b] = False
                    return True, b
            return False, 0
        
        buttonChangeState = None
        buttonNewState = None
    '''

@MethodCall(target="ButtonValue.vix", Value=int)
@Requires(buttonSetup)
def buttonValue():
    return '''
        Value = buttonValue()
    '''

@MethodCall(target="ButtonCompare.vix", Buttons=List(int), Action=int, Result=bool, Value=int)
@Requires(buttonSetup)
def buttonCompare():
    return '''
        Result, Value = buttonCompare(Buttons, Action)
    '''

@MethodCall(target="ButtonChange.vix", Result=bool, Value=int)
@Requires(buttonSetup)
def buttonChange():
    return '''
        if buttonChangeState is None:
            buttonChangeState = set(brickButtons.buttons_pressed)
        buttonNewState = set(brickButtons.buttons_pressed)
        if buttonChangeState != buttonNewState:
            Result = True
            Value = (buttonChangeState ^ buttonNewState).pop()
            buttonChangeState = None
        else:
            Result = False
    '''



"""Module for a robot that stands on two wheels and uses a gyro sensor.

The robot (eg. BALANC3R) will to keep its balance and move in response to
the remote control. This code was adapted from Laurens Valk's script at
https://github.com/laurensvalk/segway.

"""
# The MIT License (MIT)
#
# Copyright (c) 2016 Laurens Valk (laurensvalk@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import time
import json
import queue
import threading
import math
import signal
from collections import deque
from ev3dev2.power import PowerSupply
from ev3dev2.motor import LargeMotor, OUTPUT_A, OUTPUT_D
from ev3dev2.sensor.lego import GyroSensor, TouchSensor
from ev3dev2.sound import Sound
from collections import OrderedDict

log = logging.getLogger(__name__)

# Constants
RAD_PER_DEG = math.pi / 180

# EV3 Platform specific constants

# For the LEGO EV3 Gyro in Rate mode, 1 unit = 1 deg/s
DEG_PER_SEC_PER_RAW_GYRO_UNIT = 1

# Express the above as the rate in rad/s per gyro unit
RAD_PER_SEC_PER_RAW_GYRO_UNIT = DEG_PER_SEC_PER_RAW_GYRO_UNIT * RAD_PER_DEG

# For the LEGO EV3 Large Motor 1 unit = 1 deg
DEG_PER_RAW_MOTOR_UNIT = 1

# Express the above as the angle in rad per motor unit
RAD_PER_RAW_MOTOR_UNIT = DEG_PER_RAW_MOTOR_UNIT * RAD_PER_DEG

# On the EV3, "1% speed" corresponds to 1.7 RPM (if speed
# control were enabled).
RPM_PER_PERCENT_SPEED = 1.7

# Convert this number to the speed in deg/s per "percent speed"
DEG_PER_SEC_PER_PERCENT_SPEED = RPM_PER_PERCENT_SPEED * 360 / 60

# Convert this number to the speed in rad/s per "percent speed"
RAD_PER_SEC_PER_PERCENT_SPEED = DEG_PER_SEC_PER_PERCENT_SPEED * RAD_PER_DEG

# Speed and steering limits
SPEED_MAX = 20
STEER_MAX = 8


class GyroBalancer(object):
    """
    Base class for a robot that stands on two wheels and uses a gyro sensor.

    Robot will keep its balance.
    """
    def __init__(self,
                 gain_gyro_angle=1700,
                 gain_gyro_rate=120,
                 gain_motor_angle=7,
                 gain_motor_angular_speed=9,
                 gain_motor_angle_error_accumulated=3,
                 power_voltage_nominal=8.0,
                 pwr_friction_offset_nom=3,
                 timing_loop_msec=30,
                 motor_angle_history_length=5,
                 gyro_drift_compensation_factor=0.05,
                 left_motor_port=OUTPUT_D,
                 right_motor_port=OUTPUT_A,
                 debug=False):
        """Create GyroBalancer."""
        # Gain parameters
        self.gain_gyro_angle = gain_gyro_angle
        self.gain_gyro_rate = gain_gyro_rate
        self.gain_motor_angle = gain_motor_angle
        self.gain_motor_angular_speed = gain_motor_angular_speed
        self.gain_motor_angle_error_accumulated =\
            gain_motor_angle_error_accumulated

        # Power parameters
        self.power_voltage_nominal = power_voltage_nominal
        self.pwr_friction_offset_nom = pwr_friction_offset_nom

        # Timing parameters
        self.timing_loop_msec = timing_loop_msec
        self.motor_angle_history_length = motor_angle_history_length
        self.gyro_drift_compensation_factor = gyro_drift_compensation_factor

        # Power supply setup
        self.power_supply = PowerSupply()

        # Gyro Sensor setup
        self.gyro = GyroSensor()
        self.gyro.mode = self.gyro.MODE_GYRO_RATE

        # Touch Sensor setup
        self.touch = TouchSensor()

        # IR Buttons setup
        # self.remote = InfraredSensor()
        # self.remote.mode = self.remote.MODE_IR_REMOTE

        # Configure the motors
        self.motor_left = LargeMotor(left_motor_port)
        self.motor_right = LargeMotor(right_motor_port)

        # Sound setup
        self.sound = Sound()

        # Open sensor and motor files
        self.gyro_file = open(self.gyro._path + "/value0", "rb")
        self.touch_file = open(self.touch._path + "/value0", "rb")
        self.encoder_left_file = open(self.motor_left._path + "/position", "rb")
        self.encoder_right_file = open(self.motor_right._path + "/position", "rb")
        self.dc_left_file = open(self.motor_left._path + "/duty_cycle_sp", "w")
        self.dc_right_file = open(self.motor_right._path + "/duty_cycle_sp", "w")

        # Drive queue
        self.drive_queue = queue.Queue()

        # Stop event for balance thread
        self.stop_balance = threading.Event()

        # Debugging
        self.debug = debug

        # Handlers for SIGINT and SIGTERM
        signal.signal(signal.SIGINT, self.signal_int_handler)
        signal.signal(signal.SIGTERM, self.signal_term_handler)

    def shutdown(self):
        """Close all file handles and stop all motors."""
        self.stop_balance.set()  # Stop balance thread
        self.motor_left.stop()
        self.motor_right.stop()
        self.gyro_file.close()
        self.touch_file.close()
        self.encoder_left_file.close()
        self.encoder_right_file.close()
        self.dc_left_file.close()
        self.dc_right_file.close()

    def _fast_read(self, infile):
        """Function for fast reading from sensor files."""
        infile.seek(0)
        return (int(infile.read().decode().strip()))

    def _fast_write(self, outfile, value):
        """Function for fast writing to motor files."""
        outfile.truncate(0)
        outfile.write(str(int(value)))
        outfile.flush()

    def _set_duty(self, motor_duty_file, duty, friction_offset, voltage_comp):
        """Function to set the duty cycle of the motors."""
        # Compensate for nominal voltage and round the input
        duty_int = int(round(duty * voltage_comp))

        # Add or subtract offset and clamp the value between -100 and 100
        if duty_int > 0:
            duty_int = min(100, duty_int + friction_offset)
        elif duty_int < 0:
            duty_int = max(-100, duty_int - friction_offset)

        # Apply the signal to the motor
        self._fast_write(motor_duty_file, duty_int)

    def signal_int_handler(self, signum, frame):
        """Signal handler for SIGINT."""
        log.info('"Caught SIGINT')
        self.shutdown()
        raise GracefulShutdown()

    def signal_term_handler(self, signum, frame):
        """Signal handler for SIGTERM."""
        log.info('"Caught SIGTERM')
        self.shutdown()
        raise GracefulShutdown()

    def balance(self):
        """Run the _balance method as a thread."""
        balance_thread = threading.Thread(target=self._balance)
        balance_thread.start()

    def _balance(self):
        """Make the robot balance."""
        while True and not self.stop_balance.is_set():

            # Reset the motors
            self.motor_left.reset()  # Reset the encoder
            self.motor_right.reset()
            self.motor_left.run_direct()  # Set to run direct mode
            self.motor_right.run_direct()

            # Initialize variables representing physical signals
            # (more info on these in the docs)

            # The angle of "the motor", measured in raw units,
            # degrees for the EV3).
            # We will take the average of both motor positions as
            # "the motor" angle, which is essentially how far the middle
            # of the robot has travelled.
            motor_angle_raw = 0

            # The angle of the motor, converted to RAD (2*pi RAD
            # equals 360 degrees).
            motor_angle = 0

            # The reference angle of the motor. The robot will attempt to
            # drive forward or backward, such that its measured position
            motor_angle_ref = 0
            # equals this reference (or close enough).

            # The error: the deviation of the measured motor angle from the
            # reference. The robot attempts to make this zero, by driving
            # toward the reference.
            motor_angle_error = 0

            # We add up all of the motor angle error in time. If this value
            # gets out of hand, we can use it to drive the robot back to
            # the reference position a bit quicker.
            motor_angle_error_acc = 0

            # The motor speed, estimated by how far the motor has turned in
            # a given amount of time.
            motor_angular_speed = 0

            # The reference speed during manouvers: how fast we would like
            # to drive, measured in RAD per second.
            motor_angular_speed_ref = 0

            # The error: the deviation of the motor speed from the
            # reference speed.
            motor_angular_speed_error = 0

            # The 'voltage' signal we send to the motor.
            # We calculate a new value each time, just right to keep the
            # robot upright.
            motor_duty_cycle = 0

            # The raw value from the gyro sensor in rate mode.
            gyro_rate_raw = 0

            # The angular rate of the robot (how fast it is falling forward
            # or backward), measured in RAD per second.
            gyro_rate = 0

            # The gyro doesn't measure the angle of the robot, but we can
            # estimate this angle by keeping track of the gyro_rate value
            # in time.
            gyro_est_angle = 0

            # Over time, the gyro rate value can drift. This causes the
            # sensor to think it is moving even when it is perfectly still.
            # We keep track of this offset.
            gyro_offset = 0

            # Start
            log.info("Hold robot upright. Press touch sensor to start.")
            self.sound.speak("Press touch sensor to start.")

            self.touch.wait_for_bump()

            # Read battery voltage
            voltage_idle = self.power_supply.measured_volts
            voltage_comp = self.power_voltage_nominal / voltage_idle

            # Offset to limit friction deadlock
            friction_offset = int(round(self.pwr_friction_offset_nom * voltage_comp))

            # Timing settings for the program
            # Time of each loop, measured in seconds.
            loop_time_target = self.timing_loop_msec / 1000
            loop_count = 0  # Loop counter, starting at 0

            # A deque (a fifo array) which we'll use to keep track of
            # previous motor positions, which we can use to calculate the
            # rate of change (speed)
            motor_angle_hist =\
                deque([0], self.motor_angle_history_length)

            # The rate at which we'll update the gyro offset (precise
            # definition given in docs)
            gyro_drift_comp_rate =\
                self.gyro_drift_compensation_factor *\
                loop_time_target * RAD_PER_SEC_PER_RAW_GYRO_UNIT

            # Calibrate Gyro
            log.info("-----------------------------------")
            log.info("Calibrating...")

            # As you hold the robot still, determine the average sensor
            # value of 100 samples
            gyro_calibrate_count = 100
            for i in range(gyro_calibrate_count):
                gyro_offset = gyro_offset + self._fast_read(self.gyro_file)
                time.sleep(0.01)
            gyro_offset = gyro_offset / gyro_calibrate_count

            # Print the result
            log.info("gyro_offset: " + str(gyro_offset))
            log.info("-----------------------------------")
            log.info("GO!")
            log.info("-----------------------------------")
            log.info("Press Touch Sensor to re-start.")
            log.info("-----------------------------------")
            self.sound.beep()

            # Remember start time
            prog_start_time = time.time()

            if self.debug:
                # Data logging
                data = OrderedDict()
                loop_times = OrderedDict()
                data['loop_times'] = loop_times
                gyro_readings = OrderedDict()
                data['gyro_readings'] = gyro_readings

            # Initial fast read touch sensor value
            touch_pressed = False

            # Driving and Steering
            speed, steering = (0, 0)

            # Record start time of loop
            loop_start_time = time.time()

            # Balancing Loop
            while not touch_pressed and not self.stop_balance.is_set():

                loop_count += 1

                # Check for drive instructions and set speed / steering
                try:
                    speed, steering = self.drive_queue.get_nowait()
                    self.drive_queue.task_done()
                except queue.Empty:
                    pass

                # Read the touch sensor (the kill switch)
                touch_pressed = self._fast_read(self.touch_file)

                # Read the Motor Position
                motor_angle_raw = (
                    (self._fast_read(self.encoder_left_file) + self._fast_read(self.encoder_right_file)) / 2.0)
                motor_angle = motor_angle_raw * RAD_PER_RAW_MOTOR_UNIT

                # Read the Gyro
                gyro_rate_raw = self._fast_read(self.gyro_file)

                # Busy wait for the loop to reach target time length
                loop_time = 0
                while (loop_time < loop_time_target):
                    loop_time = time.time() - loop_start_time
                    time.sleep(0.001)

                # Calculate most recent loop time
                loop_time = time.time() - loop_start_time

                # Set start time of next loop
                loop_start_time = time.time()

                if self.debug:
                    # Log gyro data and loop time
                    time_of_sample = time.time() - prog_start_time
                    gyro_readings[time_of_sample] = gyro_rate_raw
                    loop_times[time_of_sample] = loop_time * 1000.0

                # Calculate gyro rate
                gyro_rate = (gyro_rate_raw - gyro_offset) *\
                    RAD_PER_SEC_PER_RAW_GYRO_UNIT

                # Calculate Motor Parameters
                motor_angular_speed_ref =\
                    speed * RAD_PER_SEC_PER_PERCENT_SPEED
                motor_angle_ref = motor_angle_ref +\
                    motor_angular_speed_ref * loop_time_target
                motor_angle_error = motor_angle - motor_angle_ref

                # Compute Motor Speed
                motor_angular_speed =\
                    ((motor_angle - motor_angle_hist[0]) /
                     (self.motor_angle_history_length * loop_time_target))
                motor_angular_speed_error = motor_angular_speed
                motor_angle_hist.append(motor_angle)

                # Compute the motor duty cycle value
                motor_duty_cycle =\
                    (self.gain_gyro_angle * gyro_est_angle +
                     self.gain_gyro_rate * gyro_rate +
                     self.gain_motor_angle * motor_angle_error +
                     self.gain_motor_angular_speed *
                     motor_angular_speed_error +
                     self.gain_motor_angle_error_accumulated *
                     motor_angle_error_acc)

                # Apply the signal to the motor, and add steering
                self._set_duty(self.dc_right_file, motor_duty_cycle + steering, friction_offset, voltage_comp)
                self._set_duty(self.dc_left_file, motor_duty_cycle - steering, friction_offset, voltage_comp)

                # Update angle estimate and gyro offset estimate
                gyro_est_angle = gyro_est_angle + gyro_rate *\
                    loop_time_target
                gyro_offset = (1 - gyro_drift_comp_rate) *\
                    gyro_offset + gyro_drift_comp_rate * gyro_rate_raw

                # Update Accumulated Motor Error
                motor_angle_error_acc = motor_angle_error_acc +\
                    motor_angle_error * loop_time_target

            # Closing down & Cleaning up

            # Loop end time, for stats
            prog_end_time = time.time()

            # Turn off the motors
            self._fast_write(self.dc_left_file, 0)
            self._fast_write(self.dc_right_file, 0)

            # Wait for the Touch Sensor to be released
            while self.touch.is_pressed:
                time.sleep(0.01)

            # Calculate loop time
            avg_loop_time = (prog_end_time - prog_start_time) / loop_count
            log.info("Loop time:" + str(avg_loop_time * 1000) + "ms")

            # Print a stop message
            log.info("-----------------------------------")
            log.info("STOP")
            log.info("-----------------------------------")

            if self.debug:
                # Dump logged data to file
                with open("data.txt", 'w') as data_file:
                    json.dump(data, data_file)

    def _move(self, speed=0, steering=0, seconds=None):
        """Move robot."""
        self.drive_queue.put((speed, steering))
        if seconds is not None:
            time.sleep(seconds)
            self.drive_queue.put((0, 0))
        self.drive_queue.join()

    def move_forward(self, seconds=None):
        """Move robot forward."""
        self._move(speed=SPEED_MAX, steering=0, seconds=seconds)

    def move_backward(self, seconds=None):
        """Move robot backward."""
        self._move(speed=-SPEED_MAX, steering=0, seconds=seconds)

    def rotate_left(self, seconds=None):
        """Rotate robot left."""
        self._move(speed=0, steering=STEER_MAX, seconds=seconds)

    def rotate_right(self, seconds=None):
        """Rotate robot right."""
        self._move(speed=0, steering=-STEER_MAX, seconds=seconds)

    def stop(self):
        """Stop robot (balancing will continue)."""
        self._move(speed=0, steering=0)


class GracefulShutdown(Exception):
    """Custom exception for SIGINT and SIGTERM."""

    pass



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
from ev3dev2 import Device

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')


class PowerSupply(Device):
    """
    A generic interface to read data from the system's power_supply class.
    Uses the built-in legoev3-battery if none is specified.
    """

    SYSTEM_CLASS_NAME = 'power_supply'
    SYSTEM_DEVICE_NAME_CONVENTION = '*'
    __slots__ = [
        '_measured_current',
        '_measured_voltage',
        '_max_voltage',
        '_min_voltage',
        '_technology',
        '_type',
    ]

    def __init__(self, address=None, name_pattern=SYSTEM_DEVICE_NAME_CONVENTION, name_exact=False, **kwargs):

        if address is not None:
            kwargs['address'] = address
        super(PowerSupply, self).__init__(self.SYSTEM_CLASS_NAME, name_pattern, name_exact, **kwargs)

        self._measured_current = None
        self._measured_voltage = None
        self._max_voltage = None
        self._min_voltage = None
        self._technology = None
        self._type = None

    @property
    def measured_current(self):
        """
        The measured current that the battery is supplying (in microamps)
        """
        self._measured_current, value = self.get_attr_int(self._measured_current, 'current_now')
        return value

    @property
    def measured_voltage(self):
        """
        The measured voltage that the battery is supplying (in microvolts)
        """
        self._measured_voltage, value = self.get_attr_int(self._measured_voltage, 'voltage_now')
        return value

    @property
    def max_voltage(self):
        self._max_voltage, value = self.get_attr_int(self._max_voltage, 'voltage_max_design')
        return value

    @property
    def min_voltage(self):
        self._min_voltage, value = self.get_attr_int(self._min_voltage, 'voltage_min_design')
        return value

    @property
    def technology(self):
        self._technology, value = self.get_attr_string(self._technology, 'technology')
        return value

    @property
    def type(self):
        self._type, value = self.get_attr_string(self._type, 'type')
        return value

    @property
    def measured_amps(self):
        """
        The measured current that the battery is supplying (in amps)
        """
        return self.measured_current / 1e6

    @property
    def measured_volts(self):
        """
        The measured voltage that the battery is supplying (in volts)
        """
        return self.measured_voltage / 1e6



# -----------------------------------------------------------------------------
# Copyright (c) 2015 Ralph Hempel <rhempel@hempeldesigngroup.com>
# Copyright (c) 2015 Anton Vanhoucke <antonvh@gmail.com>
# Copyright (c) 2015 Denis Demidov <dennis.demidov@gmail.com>
# Copyright (c) 2015 Eric Pascual <eric@pobot.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# -----------------------------------------------------------------------------

import sys
import os
import mmap
import ctypes
import logging
from PIL import Image, ImageDraw
from . import fonts
from . import get_current_platform, library_load_warning_message
from struct import pack

if sys.version_info < (3, 4):
    raise SystemError('Must be using Python 3.4 or higher')

log = logging.getLogger(__name__)

try:
    # This is a linux-specific module.
    # It is required by the Display class, but failure to import it may be
    # safely ignored if one just needs to run API tests on Windows.
    import fcntl
except ImportError:
    log.warning(library_load_warning_message("fcntl", "Display"))


class FbMem(object):
    """The framebuffer memory object.

    Made of:
        - the framebuffer file descriptor
        - the fix screen info struct
        - the var screen info struct
        - the mapped memory
    """

    # ------------------------------------------------------------------
    # The code is adapted from
    # https://github.com/LinkCareServices/cairotft/blob/master/cairotft/linuxfb.py
    #
    # The original code came with the following license:
    # ------------------------------------------------------------------
    # Copyright (c) 2012 Kurichan
    #
    # This program is free software. It comes without any warranty, to
    # the extent permitted by applicable law. You can redistribute it
    # and/or modify it under the terms of the Do What The Fuck You Want
    # To Public License, Version 2, as published by Sam Hocevar. See
    # http://sam.zoy.org/wtfpl/COPYING for more details.
    # ------------------------------------------------------------------

    __slots__ = ('fid', 'fix_info', 'var_info', 'mmap')

    FBIOGET_VSCREENINFO = 0x4600
    FBIOGET_FSCREENINFO = 0x4602

    FB_VISUAL_MONO01 = 0
    FB_VISUAL_MONO10 = 1

    class FixScreenInfo(ctypes.Structure):
        """The fb_fix_screeninfo from fb.h."""

        _fields_ = [
            ('id_name', ctypes.c_char * 16),
            ('smem_start', ctypes.c_ulong),
            ('smem_len', ctypes.c_uint32),
            ('type', ctypes.c_uint32),
            ('type_aux', ctypes.c_uint32),
            ('visual', ctypes.c_uint32),
            ('xpanstep', ctypes.c_uint16),
            ('ypanstep', ctypes.c_uint16),
            ('ywrapstep', ctypes.c_uint16),
            ('line_length', ctypes.c_uint32),
            ('mmio_start', ctypes.c_ulong),
            ('mmio_len', ctypes.c_uint32),
            ('accel', ctypes.c_uint32),
            ('reserved', ctypes.c_uint16 * 3),
        ]

    class VarScreenInfo(ctypes.Structure):
        class FbBitField(ctypes.Structure):
            """The fb_bitfield struct from fb.h."""

            _fields_ = [
                ('offset', ctypes.c_uint32),
                ('length', ctypes.c_uint32),
                ('msb_right', ctypes.c_uint32),
            ]

            def __str__(self):
                return "%s (offset %s, length %s, msg_right %s)" %\
                    (self.__class__.__name__, self.offset, self.length, self.msb_right)

        """The fb_var_screeninfo struct from fb.h."""

        _fields_ = [
            ('xres', ctypes.c_uint32),
            ('yres', ctypes.c_uint32),
            ('xres_virtual', ctypes.c_uint32),
            ('yres_virtual', ctypes.c_uint32),
            ('xoffset', ctypes.c_uint32),
            ('yoffset', ctypes.c_uint32),
            ('bits_per_pixel', ctypes.c_uint32),
            ('grayscale', ctypes.c_uint32),
            ('red', FbBitField),
            ('green', FbBitField),
            ('blue', FbBitField),
            ('transp', FbBitField),
        ]

        def __str__(self):
            return ("%sx%s at (%s,%s), bpp %s, grayscale %s, red %s, green %s, blue %s, transp %s" %
                    (self.xres, self.yres, self.xoffset, self.yoffset, self.bits_per_pixel, self.grayscale, self.red,
                     self.green, self.blue, self.transp))

    def __init__(self, fbdev=None):
        """Create the FbMem framebuffer memory object."""
        fid = FbMem._open_fbdev(fbdev)
        fix_info = FbMem._get_fix_info(fid)
        fbmmap = FbMem._map_fb_memory(fid, fix_info)
        self.fid = fid
        self.fix_info = fix_info
        self.var_info = FbMem._get_var_info(fid)
        self.mmap = fbmmap

    @staticmethod
    def _open_fbdev(fbdev=None):
        """Return the framebuffer file descriptor.

        Try to use the FRAMEBUFFER environment variable if fbdev is
        not given. Use '/dev/fb0' by default.
        """
        dev = fbdev or os.getenv('FRAMEBUFFER', '/dev/fb0')
        fbfid = os.open(dev, os.O_RDWR)
        return fbfid

    @staticmethod
    def _get_fix_info(fbfid):
        """Return the fix screen info from the framebuffer file descriptor."""
        fix_info = FbMem.FixScreenInfo()
        fcntl.ioctl(fbfid, FbMem.FBIOGET_FSCREENINFO, fix_info)
        return fix_info

    @staticmethod
    def _get_var_info(fbfid):
        """Return the var screen info from the framebuffer file descriptor."""
        var_info = FbMem.VarScreenInfo()
        fcntl.ioctl(fbfid, FbMem.FBIOGET_VSCREENINFO, var_info)
        return var_info

    @staticmethod
    def _map_fb_memory(fbfid, fix_info):
        """Map the framebuffer memory."""
        return mmap.mmap(fbfid, fix_info.smem_len, mmap.MAP_SHARED, mmap.PROT_READ | mmap.PROT_WRITE, offset=0)


class Display(FbMem):
    """
    A convenience wrapper for the FbMem class.
    Provides drawing functions from the python imaging library (PIL).
    """

    GRID_COLUMNS = 22
    GRID_COLUMN_PIXELS = 8
    GRID_ROWS = 12
    GRID_ROW_PIXELS = 10

    def __init__(self, desc='Display'):
        FbMem.__init__(self)

        self.platform = get_current_platform()

        if self.var_info.bits_per_pixel == 1:
            im_type = "1"

        elif self.platform == "ev3" and self.var_info.bits_per_pixel == 32:
            im_type = "L"

        elif self.var_info.bits_per_pixel == 16 or self.var_info.bits_per_pixel == 32:
            im_type = "RGB"

        else:
            raise Exception("Not supported - platform %s with bits_per_pixel %s" %
                            (self.platform, self.var_info.bits_per_pixel))

        self._img = Image.new(im_type, (self.fix_info.line_length * 8 // self.var_info.bits_per_pixel, self.yres),
                              "white")

        self._draw = ImageDraw.Draw(self._img)
        self.desc = desc

    def __str__(self):
        return self.desc

    @property
    def xres(self):
        """
        Horizontal screen resolution
        """
        return self.var_info.xres

    @property
    def yres(self):
        """
        Vertical screen resolution
        """
        return self.var_info.yres

    @property
    def shape(self):
        """
        Dimensions of the screen.
        """
        return (self.xres, self.yres)

    @property
    def draw(self):
        """
        Returns a handle to PIL.ImageDraw.Draw class associated with the screen.

        Example::

            screen.draw.rectangle((10,10,60,20), fill='black')
        """
        return self._draw

    @property
    def image(self):
        """
        Returns a handle to PIL.Image class that is backing the screen. This can
        be accessed for blitting images to the screen.

        Example::

            screen.image.paste(picture, (0, 0))
        """
        return self._img

    def clear(self):
        """
        Clears the screen
        """
        self._draw.rectangle(((0, 0), self.shape), fill="white")

    def _color565(self, r, g, b):
        """Convert red, green, blue components to a 16-bit 565 RGB value. Components
        should be values 0 to 255.
        """
        return (((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3))

    def _img_to_rgb565_bytes(self):
        pixels = [self._color565(r, g, b) for (r, g, b) in self._img.getdata()]
        return pack('H' * len(pixels), *pixels)

    def update(self):
        """
        Applies pending changes to the screen.
        Nothing will be drawn on the screen until this function is called.
        """
        if self.var_info.bits_per_pixel == 1:
            b = self._img.tobytes("raw", "1;R")
            self.mmap[:len(b)] = b

        elif self.var_info.bits_per_pixel == 16:
            self.mmap[:] = self._img_to_rgb565_bytes()

        elif self.var_info.bits_per_pixel == 32:
            self.mmap[:] = self._img.convert("RGB").tobytes("raw", "XRGB")

        else:
            raise Exception("Not supported - platform %s with bits_per_pixel %s" %
                            (self.platform, self.var_info.bits_per_pixel))

    def image_filename(self, filename, clear_screen=True, x1=0, y1=0, x2=None, y2=None):

        if clear_screen:
            self.clear()

        filename_im = Image.open(filename)

        if x2 is not None and y2 is not None:
            return self._img.paste(filename_im, (x1, y1, x2, y2))
        else:
            return self._img.paste(filename_im, (x1, y1))

    def line(self, clear_screen=True, x1=10, y1=10, x2=50, y2=50, line_color='black', width=1):
        """
        Draw a line from (x1, y1) to (x2, y2)
        """

        if clear_screen:
            self.clear()

        return self.draw.line((x1, y1, x2, y2), fill=line_color, width=width)

    def circle(self, clear_screen=True, x=50, y=50, radius=40, fill_color='black', outline_color='black'):
        """
        Draw a circle of 'radius' centered at (x, y)
        """

        if clear_screen:
            self.clear()

        x1 = x - radius
        y1 = y - radius
        x2 = x + radius
        y2 = y + radius

        return self.draw.ellipse((x1, y1, x2, y2), fill=fill_color, outline=outline_color)

    def rectangle(self, clear_screen=True, x1=10, y1=10, x2=80, y2=40, fill_color='black', outline_color='black'):
        """
        Draw a rectangle where the top left corner is at (x1, y1) and the
        bottom right corner is at (x2, y2)
        """

        if clear_screen:
            self.clear()

        return self.draw.rectangle((x1, y1, x2, y2), fill=fill_color, outline=outline_color)

    def point(self, clear_screen=True, x=10, y=10, point_color='black'):
        """
        Draw a single pixel at (x, y)
        """

        if clear_screen:
            self.clear()

        return self.draw.point((x, y), fill=point_color)

    def text_pixels(self, text, clear_screen=True, x=0, y=0, text_color='black', font=None):
        """
        Display ``text`` starting at pixel (x, y).

        The EV3 display is 178x128 pixels

        - (0, 0) would be the top left corner of the display
        - (89, 64) would be right in the middle of the display

        ``text_color`` : PIL says it supports "common HTML color names". There
        are 140 HTML color names listed here that are supported by all modern
        browsers. This is probably a good list to start with.
        https://www.w3schools.com/colors/colors_names.asp

        ``font`` : can be any font displayed here
            http://ev3dev-lang.readthedocs.io/projects/python-ev3dev/en/ev3dev-stretch/display.html#bitmap-fonts

        - If font is a string, it is the name of a font to be loaded.
        - If font is a Font object, returned from :meth:`ev3dev2.fonts.load`, then it is
          used directly.  This is desirable for faster display times.

        """

        if clear_screen:
            self.clear()

        if font is not None:
            if isinstance(font, str):
                assert font in fonts.available(), "%s is an invalid font" % font
                font = fonts.load(font)
            return self.draw.text((x, y), text, fill=text_color, font=font)
        else:
            return self.draw.text((x, y), text, fill=text_color)

    def text_grid(self, text, clear_screen=True, x=0, y=0, text_color='black', font=None):
        """
        Display ``text`` starting at grid (x, y)

        The EV3 display can be broken down in a grid that is 22 columns wide
        and 12 rows tall. Each column is 8 pixels wide and each row is 10
        pixels tall.

        ``text_color`` : PIL says it supports "common HTML color names". There
        are 140 HTML color names listed here that are supported by all modern
        browsers. This is probably a good list to start with.
        https://www.w3schools.com/colors/colors_names.asp

        ``font`` : can be any font displayed here
            http://ev3dev-lang.readthedocs.io/projects/python-ev3dev/en/ev3dev-stretch/display.html#bitmap-fonts

        - If font is a string, it is the name of a font to be loaded.
        - If font is a Font object, returned from :meth:`ev3dev2.fonts.load`, then it is
          used directly.  This is desirable for faster display times.

        """

        assert 0 <= x < Display.GRID_COLUMNS,\
            "grid columns must be between 0 and %d, %d was requested" %\
            ((Display.GRID_COLUMNS - 1, x))

        assert 0 <= y < Display.GRID_ROWS,\
            "grid rows must be between 0 and %d, %d was requested" %\
            ((Display.GRID_ROWS - 1), y)

        return self.text_pixels(text, clear_screen, x * Display.GRID_COLUMN_PIXELS, y * Display.GRID_ROW_PIXELS,
                                text_color, font)

    def reset_screen(self):
        self.clear()
        self.update()



#!/usr/bin/env python3
"""
Wheel and Rim classes

A great reference when adding new wheels is http://wheels.sariel.pl/
"""
from math import pi


class Wheel(object):
    """
    A base class for various types of wheels, tires, etc.  All units are in mm.

    One scenario where one of the child classes below would be used is when the
    user needs their robot to drive at a specific speed or drive for a specific
    distance. Both of those calculations require the circumference of the wheel
    of the robot.

    Example:

    .. code:: python

        from ev3dev2.wheel import EV3Tire

        tire = EV3Tire()

        # calculate the number of rotations needed to travel forward 500 mm
        rotations_for_500mm = 500 / tire.circumference_mm
    """
    def __init__(self, diameter_mm, width_mm):
        self.diameter_mm = float(diameter_mm)
        self.width_mm = float(width_mm)
        self.circumference_mm = diameter_mm * pi

    @property
    def radius_mm(self):
        return float(self.diameter_mm / 2)


class EV3Rim(Wheel):
    """
    part number 56145
    comes in set 31313
    """
    def __init__(self):
        Wheel.__init__(self, 30, 20)


class EV3Tire(Wheel):
    """
    part number 44309
    comes in set 31313
    """
    def __init__(self):
        Wheel.__init__(self, 43.2, 21)


class EV3EducationSetRim(Wheel):
    """
    part number 56908
    comes in set 45544
    """
    def __init__(self):
        Wheel.__init__(self, 43, 26)


class EV3EducationSetTire(Wheel):
    """
    part number 41897
    comes in set 45544
    """
    def __init__(self):
        Wheel.__init__(self, 56, 28)



from .types import Literal
from .utils import MethodCall, DynamicMethodCall, p

@MethodCall(target="CaseSelector_Boolean.vix", Boolean=bool, Result=bool)
def caseSelectorBoolean():
    return '''
        Result = Boolean
    '''

@MethodCall(target="CaseSelector_Numeric.vix", Number=int, Result=int)
def caseSelectorNumeric():
    return '''
        Result = Number
    '''

@MethodCall(target="CaseSelector_String.vix", String=str, Result=str)
def caseSelectorString():
    return '''
        Result = String
    '''

@MethodCall(target="LoopIndex.vix", LoopIterationCount=int, LoopIndex=int)
def loopIndex():
    return '''
        LoopIndex = LoopIterationCount - 1
    '''

@DynamicMethodCall(target="ToggleInterrupt.vix", InterruptName=Literal(str), LoopLabel=Literal(str))
def loopInterrupt(InterruptName, LoopLabel):
    if LoopLabel != InterruptName:
        raise Exception(f"Loop Interrupt {InterruptName} only supports breaking a loop it is directly inside")

    return p('''
        break
    ''')

@MethodCall(target="StopAfterNumberIterations.vix", IterationsToRun=int, LoopIterationCount=int, Result=bool)
def stopAfterNumberIterations():
    return '''
        Result = LoopIterationCount >= IterationsToRun
    '''

@MethodCall(target="StopIfTrue.vix", DoStop=bool, Result=bool)
def stopIfTrue():
    return '''
        Result = DoStop
    '''

@MethodCall(target="StopNever.vix", Result=bool)
def stopNever():
    return '''
        Result = False
    '''

@MethodCall(target="TimeCompareLoop.vix", HowLong=float, Result=bool, LoopEnterTime=float)
def timeCompareLoop():
    return '''
        Result = time.time() >= LoopEnterTime + HowLong
    '''

@MethodCall(target="TimeCompare.vix", HowLong=float, Result=bool, WaitEnterTime=float)
def timeCompareWait():
    return '''
        Result = time.time() >= WaitEnterTime + HowLong
    '''



from mindpile.Mapping.utils import MethodCall

@MethodCall(target="InsideRange.vix", TestValue=float, LowerBound=float, UpperBound=float, Result=bool)
def insideRange():
    return '''
        Result = TestValue >= LowerBound and TestValue <= UpperBound
    '''

@MethodCall(target="OutsideRange.vix", TestValue=float, LowerBound=float, UpperBound=float, Result=bool)
def outsideRange():
    return '''
        Result = TestValue < LowerBound and TestValue > UpperBound
    '''



from mindpile.Mapping.types import List
from mindpile.Mapping.utils import MethodCall

@MethodCall(target="ArrayBuild.vix", arrayInNumeric=List(float), valueIn=float, arrayOutNumeric=List(float))
def arrayAppendNumeric():
    return '''
        arrayOutNumeric = [*arrayInNumeric, valueIn]
    '''

@MethodCall(target="ArrayBuildBoolean.vix", arrayInBoolean=List(bool), valueIn=bool, arrayOutBoolean=List(bool))
def arrayAppendBoolean():
    return '''
        arrayOutBoolean = [*arrayInBoolean, valueIn]
    '''

@MethodCall(target="ArrayReadAtIndex.vix", arrayInNumeric=List(float), Index=int, valueOut=float)
def arrayReadNumeric():
    return '''
        valueOut = arrayInNumeric[Index]
    '''

@MethodCall(target="ArrayReadAtIndexBoolean.vix", arrayInBoolean=List(bool), Index=int, valueOut=bool)
def arrayReadBoolean():
    return '''
        valueOut = arrayInBoolean[Index]
    '''

@MethodCall(target="ArrayWriteAtIndex.vix", arrayInNumeric=List(float), Index=int, valueIn=float, arrayOutNumeric=List(float))
def arrayWriteNumeric():
    return '''
        arrayOutNumeric = list(arrayInNumeric)
        arrayOutNumeric[Index] = valueIn
    '''

@MethodCall(target="ArrayWriteAtIndexBoolean.vix", arrayInBoolean=List(bool), Index=int, valueIn=bool, arrayOutBoolean=List(bool))
def arrayWriteBoolean():
    return '''
        arrayOutBoolean = list(arrayInBoolean)
        arrayOutBoolean[Index] = valueIn
    '''

@MethodCall(target="ArrayGetSize.vix", arrayInNumeric=List(float), Size=int)
def arrayLengthNumeric():
    return '''
        Size = len(arrayInNumeric)
    '''

@MethodCall(target="ArrayGetSizeBoolean.vix", arrayInBoolean=List(bool), Size=int)
def arrayLengthBoolean():
    return '''
        Size = len(arrayInBoolean)
    '''
